; A 6502 emulator
; /Mic, 2004


.486
.model flat,stdcall
option casemap:none


include \masm32\include\windows.inc
include \masm32\include\kernel32.inc
include \masm32\include\user32.inc

includelib \masm32\lib\kernel32.lib
includelib \masm32\lib\user32.lib
wsprintfA PROTO C :DWORD, :VARARG
wsprintf equ <wsprintfA>

.data

	CPU_IRQ_CALLBACK		equ 000h
	CPU_SET_IO_CALLBACK		equ 001h
	CPU_IO_HANDLER			equ 001h
	CPU_CYCLE_COUNT			equ 002h
	CPU_REG_PC			equ 003h
	CPU_REG_SP			equ 004h
	CPU_BREAKPOINT			equ 005h
	CPU_SPEED			equ 006h
	CPU_INPB_CALLBACK		equ 007h
	CPU_OUTPB_CALLBACK		equ 008h
	CPU_CYCLE_SCALING		equ 009h
	
	MMU_READ_BYTE_CALLBACK		equ 100h
	MMU_READ_WORD_CALLBACK		equ 101h
	MMU_WRITE_BYTE_CALLBACK		equ 102h
	MMU_WRITE_WORD_CALLBACK		equ 103h
	MMU_TRANSLATE_ADDRESS_CALLBACK	equ 104h

	GPU_CURRENT_SCANLINE		equ 200h
	GPU_SCREEN_WIDTH		equ 201h
	GPU_SCREEN_HEIGHT		equ 202h
	GPU_FRAME_COMPLETED		equ 203h
	GPU_SCREEN_PITCH		equ 204h

	ZOMBIE_WINDOW_HANDLE		equ 500h

	PLUGIN_NAME			equ 600h
	
	INPUT		equ 0
	OUTPUT		equ 1

	ASK		equ 0
	TELL		equ 1

	; 6502 flag masks	
	C_FLAG 	equ 01h		; carry
	Z_FLAG	equ 02h		; zero
	I_FLAG	equ 04h		; interrupt 
	D_FLAG	equ 08h		; decimal
	B_FLAG	equ 10h		; break
	V_FLAG	equ 40h		; overflow	
	S_FLAG	equ 80h		; sign


	; Various prototypes
	_1ARGS_PROTO	typedef PROTO :DWORD
	_2ARGS_PROTO	typedef PROTO :DWORD,:DWORD
	_4ARGS_PROTO	typedef PROTO :DWORD,:DWORD,:DWORD,:DWORD
	_1ARGS		typedef ptr _1ARGS_PROTO
	_2ARGS		typedef ptr _2ARGS_PROTO
	_4ARGS		typedef ptr _4ARGS_PROTO


	cpu_inpb	PROTO :DWORD
	cpu_outpb	PROTO :DWORD,:DWORD


start_regs label byte
	_A	db 0
	_X	db 0
	_Y	db 0
	_S	db 0
	_P	db 0
	_PC	dw 0
	

	Carry	db 0
	Zero	db 0
	Interrupt db 0
	Decimal	db 0
	Break	db 0
	Overflow db 0
	Sign	db 0
	
	cycle	dd 0

	lastPC	dw 0
	breakpoint dd 80000000h
end_regs label byte

	halted	dd 0
	speed	dd 0380000h		; default speed: 3.5 MHz
	cpuName db "6502",0


	szBuffer	db 80 DUP(0)
	szError		db "Error",0
	szFormat db "Setting io handler for port %02x, mode %d",13,10,0
	szFormat2 db "Reading from port %02x.",13,10,0
	szFormat3 db "IO handler = %08x, null_handler = %08x",13,10,0
	szFormat4 db "Read %02x from %04x",13,10,0
	szFormat5 db "PC=0000. last good PC=%04x",13,10,0
	hout dd 0


.data?
	hInst	dd ?
	RAM	dd ?
	dummy dd ?
	zombie	_4ARGS ?
	mmu_read_byte _1ARGS ?
	mmu_read_word _1ARGS ?
	mmu_write_byte _2ARGS ?
	mmu_write_word _2ARGS ?
	mmu_get _1ARGS ?
	mmu_translate_address _1ARGS ?
	
.code





ADD_CYCLES MACRO num
;	add ecx,num*400h
	add cycle,num*400h
ENDM


; Update the Sign and Zero flags
UPDATE_NZ MACRO
	sets Sign
	setz Zero
ENDM


; Update the Sign, Zero and Carry flags
UPDATE_NZC MACRO
	sets Sign
	setz Zero
	setc Carry
ENDM


; Update the Sign, Zero, Overflow and Carry flags
UPDATE_NZVC MACRO
	sets Sign
	setz Zero
	seto Overflow
	setc Carry
ENDM


; Do a relative branch 
BRANCH MACRO flag,jump
	LOCAL label
	cmp flag,0
	jump label
		movsx eax,byte ptr [esi]
		add ax,2
		add _PC,ax
		
		mov ax,_PC
		mov bx,lastPC
		and eax,0FF00h
		and ebx,0FF00h
		.if eax != ebx
			ADD_CYCLES 4
		.else
			ADD_CYCLES 3
		.endif
		ret
	label:	
		ADD_CYCLES 3
		add _PC,2
		ret
ENDM


; Load an immediate value into reg
LDIMM MACRO reg
	movzx eax,byte ptr [esi]
	and eax,eax
	UPDATE_NZ
	add _PC,2
	ADD_CYCLES 3
	mov reg,al
	ret
ENDM


; Load a value from the zeropage into reg
LDZP MACRO reg
	movzx ebx,byte ptr [esi]
	invoke mmu_read_byte,ebx
	and eax,eax
	UPDATE_NZ
	add _PC,2
	ADD_CYCLES 3
	mov reg,al
	ret
ENDM


; Load a value from the zeropage with index idx into reg
LDZPX MACRO reg,idx
	movzx ebx,byte ptr [esi]
	add bl,idx
	invoke mmu_read_byte,ebx
	and eax,eax
	UPDATE_NZ
	add _PC,2
	ADD_CYCLES 4
	mov reg,al
	ret
ENDM


LDABS MACRO reg
	movzx ebx,word ptr [esi]
	invoke mmu_read_byte,ebx
	and eax,eax
	UPDATE_NZ
	add _PC,3
	ADD_CYCLES 4
	mov reg,al
	ret
ENDM


LDABSX MACRO reg,idx
	movzx eax,byte ptr idx
	movzx ebx,word ptr [esi]
	add bx,ax
	invoke mmu_read_byte,ebx
	and eax,eax
	UPDATE_NZ
	add _PC,3
	ADD_CYCLES 4
	mov reg,al
	ret
ENDM


RDBABS MACRO
	movzx ebx,word ptr [esi]
	invoke mmu_read_byte,ebx
ENDM


RDBABSX MACRO idx
	movzx eax,byte ptr idx
	movzx ebx,word ptr [esi]
	add bx,ax
	invoke mmu_read_byte,ebx
ENDM


RDBIMM MACRO
	movzx eax,byte ptr [esi]
ENDM


RDBZP MACRO
	movzx ebx,byte ptr [esi]
	invoke mmu_read_byte,ebx
ENDM


RDBZPX MACRO idx
	movzx ebx,byte ptr [esi]
	add bl,idx
	invoke mmu_read_byte,ebx
ENDM


RDBINX MACRO 
	movzx ebx,byte ptr [esi]
	add bl,_X
	invoke mmu_read_word,ebx
	and eax,0FFFFh
	invoke mmu_read_byte,eax
ENDM


RDBINY MACRO 
	movzx ebx,byte ptr [esi]
	invoke mmu_read_word,ebx
	movzx ebx,byte ptr _Y
	and eax,0FFFFh
	add bx,ax
	invoke mmu_read_byte,ebx
ENDM


WRBABS MACRO
	movzx ebx,word ptr [esi]
	invoke mmu_write_byte,ebx,eax
ENDM


WRBZP MACRO
	movzx ebx,byte ptr [esi]
	invoke mmu_write_byte,ebx,eax
ENDM


ADC_ MACRO
	cmp Carry,1
	cmc
	adc _A,al
	UPDATE_NZVC
ENDM


AND_ MACRO
	and _A,al
	UPDATE_NZ
ENDM


ASL_ MACRO
	shl al,1
	UPDATE_NZC
ENDM


BIT_ MACRO
	test al,80h
	setnz Sign
	test al,40h
	setnz Overflow
	and al,_A
	setz Zero
ENDM


CPR MACRO reg
	cmp reg,al
	UPDATE_NZC
ENDM


DEC_ MACRO
	dec al
	UPDATE_NZ
ENDM


EOR_ MACRO
	xor _A,al
	UPDATE_NZ
ENDM


INC_ MACRO
	inc al
	UPDATE_NZ
ENDM


ORA_ MACRO
	or _A,al
	UPDATE_NZ
ENDM


ROL_ MACRO
	cmp Carry,1
	cmc
	rcl al,1
	UPDATE_NZC
ENDM


SBC_ MACRO
	cmp Carry,1
	sbb _A,al
	UPDATE_NZVC
ENDM


ABSOPTAIL MACRO
	ADD_CYCLES 4
	add _PC,3
	ret
ENDM


IMMOPTAIL MACRO
	ADD_CYCLES 2
	add _PC,2
	ret
ENDM


ZPOPTAIL MACRO
	ADD_CYCLES 3
	add _PC,2
	ret
ENDM


STZP MACRO reg
	movzx ebx,byte ptr [esi]
	movzx eax,reg
	invoke mmu_write_byte,ebx,eax
	add _PC,2
	ADD_CYCLES 3
	ret
ENDM


STZPX MACRO reg,idx
	movzx ebx,byte ptr [esi]
	movzx eax,reg
	add bl,idx
	invoke mmu_write_byte,ebx,eax
	add _PC,2
	ADD_CYCLES 4
	ret
ENDM


STABS MACRO reg
	movzx ebx,word ptr [esi]
	movzx eax,reg
	invoke mmu_write_byte,ebx,eax
	add _PC,3
	ADD_CYCLES 4
	ret
ENDM


STABSX MACRO reg,idx
	movzx eax,byte ptr idx
	movzx ebx,word ptr [esi]
	add bx,ax
	movzx eax,reg
	invoke mmu_write_byte,ebx,eax
	add _PC,3
	ADD_CYCLES 5
	ret
ENDM


PULLB MACRO 
	inc _S
	movzx eax,byte ptr _S
	add eax,100h
	invoke mmu_read_byte,eax
	and eax,0FFh
ENDM


PULLW MACRO 
	inc _S
	movzx eax,byte ptr _S
	add eax,100h
	inc _S
	invoke mmu_read_byte,eax
	and eax,0FFh
ENDM


PUSHB MACRO reg
	movzx ebx,byte ptr _S
	add ebx,100h
	invoke mmu_write_byte,ebx,reg
	dec _S
ENDM

PUSHWD MACRO reg
	movzx ebx,byte ptr _S
	dec ebx
	add ebx,100h
	invoke mmu_write_word,ebx,reg
	sub _S,2
ENDM
	

;-------------------------------------------------------------------------------------------------

	
LibMain PROC hInstDLL:DWORD, reason:DWORD, unused:DWORD

        .if reason == 1		;DLL_PROCESS_ATTACH
        	mov eax,hInstDLL
        	mov hInst,eax
		mov eax,TRUE
		ret

        .elseif reason == 0	;DLL_PROCESS_DETACH
        .elseif reason == 2	;DLL_THREAD_ATTACH
        .elseif reason == 3	;DLL_THREAD_DETACH
        .endif

        mov eax,FALSE
        ret
LibMain ENDP


inpb_null PROC
	ret
inpb_null ENDP
.data
inpb_handlers dd 256 DUP(OFFSET inpb_null)
.code

outpb_null PROC dat:DWORD
	ret
outpb_null ENDP
.data
outpb_handlers dd 256 DUP(OFFSET outpb_null)
.code	


ALIGN 8

; BRK
op_00:
	ret

; ORA (aa,X)	
op_01:
	RDBINX
	ORA_
	ADD_CYCLES 6
	add _PC,2
	ret
	
; Illegal	
op_02:
; Illegal
op_03:
; Illegal
op_04:
	ret

; ORA aa	
op_05:
	RDBZP
	ORA_
	ZPOPTAIL
	
; ASL aa	
op_06:
	RDBZP
	ASL_
	WRBZP
	ZPOPTAIL
	
; Illegal
op_07:
	ret
	
; PHP
op_08:
	mov al,Sign
	add eax,eax
	or al,Overflow
	add eax,eax
	add eax,eax
	or eax,3
	add eax,eax
	or al,Decimal
	add eax,eax
	or al,Interrupt
	add eax,eax
	or al,Zero
	add eax,eax
	or al,Carry
	PUSHB eax
	ADD_CYCLES 3
	inc _PC
	ret

; ORA #aa
op_09:
	RDBIMM
	ORA_
	IMMOPTAIL
	
; ASL A	
op_0A:
	shl _A,1
	UPDATE_NZC
	ADD_CYCLES 2
	inc _PC
	ret

; Illegal	
op_0B:
; Illegal
op_0C:
	ret

; ORA aaaa	
op_0D:
	RDBABS
	ORA_
	ABSOPTAIL
	
; ASL aaaa	
op_0E:
	RDBABS
	ASL_
	WRBABS
	ABSOPTAIL
	
; Illegal
op_0F:
	ret
	
; BPL
op_10:
	BRANCH Sign,jne

; ORA (aa),Y	
op_11:
	RDBINY
	ORA_
	ADD_CYCLES 5
	add _PC,2
	ret

; Illegal	
op_12:
; Illegal
op_13:
;Illegal
op_14:
	ret

; ORA aa,X	
op_15:
	RDBZPX _X
	ORA_
	add _PC,2
	ADD_CYCLES 4
	ret

; ASL aa,X
op_16:
	
; Illegal
op_17:
	ret
	
; CLC
op_18:
	mov Carry,0
	inc _PC
	ADD_CYCLES 2
	ret

; ORA aaaa,Y	
op_19:
	RDBABSX _Y
	ORA_
	ADD_CYCLES 4
	add _PC,3
	ret

; Illegal	
op_1A:
; Illegal
op_1B:
; Illegal
op_1C:
	ret

; ORA aaaa,X
op_1D:
	RDBABSX _X
	ORA_
	ADD_CYCLES 4
	add _PC,3
	ret

; ASL aaaa,X
op_1E:

; Illegal
op_1F:
	ret
	
; JSR
op_20:
	movzx eax,word ptr _PC
	dec eax
	mov bx,[esi]
	mov _PC,bx
	PUSHWD eax
	ADD_CYCLES 6
	ret

; AND (aa,X)	
op_21:
	RDBINX
	AND_
	ADD_CYCLES 6
	add _PC,2
	ret
	
; Illegal
op_22:
; Illegal
op_23:
	ret

; BIT aa	
op_24:
	RDBZP
	BIT_
	ZPOPTAIL
	
; AND aa
op_25:
	RDBZP
	AND_
	ZPOPTAIL

; ROL aa	
op_26:
	RDBZP
	ROL_
	WRBZP
	ZPOPTAIL
	
op_27:

; PLP
op_28:
	PULLB
	test eax,C_FLAG
	setne Carry
	test eax,Z_FLAG
	setne Zero
	test eax,I_FLAG
	setne Interrupt
	test eax,D_FLAG
	setne Decimal
	test eax,B_FLAG
	setne Break
	test eax,V_FLAG
	setne Overflow
	test eax,S_FLAG
	setne Sign
	ADD_CYCLES 4
	inc _PC
	ret
op_29:

; ROL A
op_2A:
	cmp Carry,1
	cmc
	rcl _A,1
	UPDATE_NZC
	ADD_CYCLES 2
	inc _PC
	ret

; Illegal	
op_2B:
	ret
	
; BIT aaaa	
op_2C:
	RDBABS
	BIT_
	ABSOPTAIL

; AND aaaa
op_2D:
	RDBABS
	AND_
	ABSOPTAIL

; ROL aaaa
op_2E:
	RDBABS
	ROL_
	WRBABS
	ADD_CYCLES 6
	add _PC,3
	ret
	
; Illegal
op_2F:
	ret
	
; BMI
op_30:
	BRANCH Sign,je

; AND (aa),Y	
op_31:
	RDBINY
	AND_
	ADD_CYCLES 5
	add _PC,2
	ret

; Illegal
op_32:
; Illegal
op_33:
; Illegal
op_34:
	ret

; AND aa,X	
op_35:
	RDBZPX _X
	AND_
	ADD_CYCLES 4
	add _PC,2
	ret

; ROL aa,X	
op_36:

; Illegal
op_37:
	ret
; SEC
op_38:
	mov Carry,1
	ADD_CYCLES 2
	inc _PC
	ret

; AND aaaa,Y	
op_39:
	RDBABSX _Y
	AND_
	ADD_CYCLES 4
	add _PC,3
	ret
	
; Illegal
op_3A:
; Illegal
op_3B:
; Illegal
op_3C:
	ret
	
; AND aaaa,X	
op_3D:
	RDBABSX _X
	AND_
	ADD_CYCLES 4
	add _PC,3
	ret

; ROL aaaa,X	
op_3E:

; Illegal
op_3F:
	ret
	
; RTI
op_40:
	PULLB
	test eax,C_FLAG
	setne Carry
	test eax,Z_FLAG
	setne Zero
	test eax,I_FLAG
	setne Interrupt
	test eax,D_FLAG
	setne Decimal
	test eax,B_FLAG
	setne Break
	test eax,V_FLAG
	setne Overflow
	test eax,S_FLAG
	setne Sign

	PULLW
	mov _PC,ax
	ADD_CYCLES 6
	ret

; EOR (aa,X)	
op_41:
	RDBINX
	EOR_
	ADD_CYCLES 6
	add _PC,2
	ret
	
; Illegal
op_42:
; Illegal
op_43:
; Illegal
op_44:
	ret
	
; EOR aa
op_45:
	RDBZP
	EOR_
	ZPOPTAIL
	
op_46:

; Illegal
op_47:
	ret
	
; PHA
op_48:
	mov al,_A
	PUSHB eax
	ADD_CYCLES 3
	inc _PC
	ret

; EOR #aa	
op_49:
	RDBIMM
	EOR_
	ADD_CYCLES 2
	add _PC,2
	ret

; LSR A	
op_4A:
	shr _A,1
	UPDATE_NZC
	ADD_CYCLES 2
	inc _PC
	ret

; Illegal	
op_4B:
	ret
	
; JMP aaaa
op_4C:
	movzx eax,word ptr [esi]
	ADD_CYCLES 3
	mov _PC,ax
	ret
	
op_4D:
op_4E:

; Illegal
op_4F:
	ret
	
op_50:
op_51:
op_52:
op_53:
op_54:

; EOR aa,X
op_55:
	RDBZPX _X
	EOR_
	ADD_CYCLES 4
	add _PC,2
	ret
	
op_56:
op_57:

; CLI
op_58:
	mov Interrupt,0
	ADD_CYCLES 2
	inc _PC
	ret
	
op_59:
op_5A:
op_5B:
op_5C:
op_5D:
op_5E:
op_5F:

; RTS
op_60:
	PULLW
	inc ax

	mov _PC,ax
	ADD_CYCLES 6
	ret
	
op_61:
op_62:
op_63:
op_64:

; ADC aa
op_65:
	RDBZP
	ADC_
	ADD_CYCLES 3
	add _PC,2
	ret
	
op_66:
op_67:

; PLA
op_68:
	PULLB
	and eax,0FFh
	UPDATE_NZ
	mov _A,al
	ADD_CYCLES 4
	inc _PC

	ret
	
; ADC #aa
op_69:
	RDBIMM
	ADC_
	ADD_CYCLES 2
	add _PC,2
	ret

; ROR A
op_6A:
	cmp Carry,1
	cmc
	rcr _A,1
	UPDATE_NZC
	ADD_CYCLES 2
	inc _PC
	ret

; Illegal
op_6B:
	ret

; JMP (aaaa)	
op_6C:

; ADC aaaa
op_6D:
	RDBABS
	ADC_
	ADD_CYCLES 4
	add _PC,3
	ret
	
op_6E:

; Illegal
op_6F:
	ret
	
op_70:
op_71:
op_72:
op_73:
op_74:

; ADC aa,X
op_75:
	RDBZPX _X
	ADC_
	ADD_CYCLES 2
	add _PC,4
	ret
	
op_76:
op_77:

; SEI
op_78:
	mov Interrupt,1
	ADD_CYCLES 2
	inc _PC
	ret
	
op_79:
op_7A:
op_7B:
op_7C:
op_7D:
op_7E:
op_7F:

op_80:
op_81:
op_82:
op_83:

; STY aa
op_84:
	STZP _Y

; STA aa
op_85:
	STZP _A

; STX aa	
op_86:
	STZP _X

op_87:

; DEY
op_88:
	dec byte ptr _Y
	UPDATE_NZ
	inc _PC
	ADD_CYCLES 2
	ret
op_89:

; TXA
op_8A:
	mov al,_X
	or al,al
	UPDATE_NZ
	mov _A,al
	ADD_CYCLES 2
	inc _PC
	ret
	
op_8B:

; STY aaaa
op_8C:
	STABS _Y
	
; STA aaaa
op_8D:
	STABS _A

; STX aaaa	
op_8E:
	STABS _X
	
op_8F:

; BCC
op_90:
	BRANCH Carry,jne
op_91:
op_92:
op_93:

; STY aa,X
op_94:
	STZPX _Y,_X
; STA aa,X
op_95:
	STZPX _A,_X
; STX aa,Y
op_96:
	STZPX _X,_Y

; Illegal	
op_97:
	ret
	
; TYA
op_98:
	mov al,_Y
	or al,al
	UPDATE_NZ
	mov _A,al
	ADD_CYCLES 2
	inc _PC
	ret

; STA aaaa,Y	
op_99:
	STABSX _A,_Y
	
; TXS
op_9A:
	mov al,_X
	mov _S,al
	ADD_CYCLES 2
	inc _PC
	ret

; Illegal	
op_9B:
; Illegal
op_9C:

; STA aaaa,X
op_9D:
	STABSX _A,_X

; Illegal	
op_9E:
; Illegal
op_9F:
	ret
	
; LDY #aa
op_A0:
	LDIMM _Y

op_A1:

; LDX #aa
op_A2:
	LDIMM _X

op_A3:

; LDY aa
op_A4:
	LDZP _Y

; LDA aa
op_A5:
	LDZP _A

; LDX aa
op_A6:
	LDZP _X

op_A7:

; TAY
op_A8:
	mov al,_A
	or al,al
	UPDATE_NZ
	mov _Y,al
	ADD_CYCLES 2
	inc _PC
	ret
	
; LDA #aa
op_A9:
	LDIMM _A

; TAX	
op_AA:
	mov al,_A
	or al,al
	UPDATE_NZ
	mov _X,al
	ADD_CYCLES 2
	inc _PC
	ret
op_AB:
op_AC:

; LDA aaaa
op_AD:
	LDABS _A
	
op_AE:
op_AF:

op_B0:
op_B1:
op_B2:
op_B3:

; LDY aa,X
op_B4:
	LDZPX _Y,_X

; LDA aa,X
op_B5:
	LDZPX _A,_X
	
; LDX aa,Y
op_B6:
	LDZPX _X,_Y
	
op_B7:

; CLV
op_B8:
	mov Overflow,0
	ADD_CYCLES 2
	inc _PC
	ret
	
op_B9:

; TSX
op_BA:
	mov al,_S
	or al,al
	UPDATE_NZ
	mov _X,al
	ADD_CYCLES 2
	inc _PC
	ret

; Illegal	
op_BB:
	ret
	
; LDY aaaa,X
op_BC:
	LDABSX _Y,_X
	
; LDA aaaa,X
op_BD:
	LDABSX _A,_X

; LDX aaaa,Y	
op_BE:
	LDABSX _X,_Y
	
; Illegal
op_BF:
	ret
	
; CPY #aa
op_C0:
	RDBIMM
	CPR _Y
	ADD_CYCLES 2
	add _PC,2
	ret
	
; CMP (aa,X)	
op_C1:
	RDBINX
	CPR _A
	ADD_CYCLES 6
	add _PC,2
	ret

; Illegal	
op_C2:
; Illegal
op_C3:
	ret
	
; CPY aa
op_C4:
	RDBZP
	CPR _Y
	ADD_CYCLES 3
	add _PC,2
	ret
	
; CMP aa 
op_C5:
	RDBZP
	CPR _A
	ADD_CYCLES 3
	add _PC,2
	ret

; DEC aa	
op_C6:
	RDBZP
	DEC_
	WRBZP
	ADD_CYCLES 5
	add _PC,2
	ret
	
; Illegal
op_C7:
	ret
	
; INY
op_C8:
	inc byte ptr _Y
	UPDATE_NZ
	inc _PC
	ADD_CYCLES 2
	ret

; CMP #aa	
op_C9:
	RDBIMM
	CPR _A
	ADD_CYCLES 2
	add _PC,2
	ret

; DEX
op_CA:
	dec byte ptr _X
	UPDATE_NZ
	inc _PC
	ADD_CYCLES 2
	ret
op_CB:

; CPY aaaa
op_CC:
	RDBABS
	CPR _Y
	ADD_CYCLES 4
	add _PC,3
	ret

; CMP aaaa
op_CD:
	RDBABS
	CPR _A
	ADD_CYCLES 4
	add _PC,3
	ret

; DEC aaaa	
op_CE:
	RDBABS
	DEC_
	WRBABS
	ADD_CYCLES 6
	add _PC,3
	ret

; Illegal	
op_CF:
	ret
	
; BNE aa
op_D0:
	BRANCH Zero,jne

; CMP (aa),Y	
op_D1:
	RDBINY
	CPR _A
	ADD_CYCLES 5
	add _PC,2
	ret
	
; Illegal
op_D2:
; Illegal
op_D3:
; illegal
op_D4:
	ret
	
; CMP aa,X
op_D5:
	RDBZPX _X
	CPR _A
	ADD_CYCLES 4
	add _PC,2
	ret
op_D6:
op_D7:
op_D8:
op_D9:

; Illegal
op_DA:
; Illegal
op_DB:
; Illegal
op_DC:
	ret
	
op_DD:
op_DE:
op_DF:

; CPX #aa
op_E0:
	RDBIMM
	CPR _X
	ADD_CYCLES 2
	add _PC,2
	ret
	
op_E1:
op_E2:
op_E3:

; CPX aa
op_E4:
	RDBZP
	CPR _X
	ADD_CYCLES 3
	add _PC,2
	ret

; SBC aa
op_E5:
	RDBZP
	SBC_
	ADD_CYCLES 3
	add _PC,2
	ret

; INC aa	
op_E6:
	RDBZP
	INC_
	WRBZP
	ADD_CYCLES 5
	add _PC,2
	ret
	
op_E7:

; INX
op_E8:
	dec byte ptr _X
	UPDATE_NZ
	inc _PC
	ADD_CYCLES 2
	ret

; SBC #aa
op_E9:
	RDBIMM
	ADC_
	ADD_CYCLES 2
	add _PC,2
	ret

; NOP	
op_EA:
	ADD_CYCLES 2
	inc _PC
	ret

; Illegal	
op_EB:
	ret
	
; CPX aaaa
op_EC:
	RDBABS
	CPR _X
	ADD_CYCLES 4
	add _PC,3
	ret
op_ED:

; INC aaaa
op_EE:
	RDBABS
	INC_
	WRBABS
	ADD_CYCLES 6
	add _PC,3
	ret

; Illegal	
op_EF:
	ret
	
; BEQ aa
op_F0:
	BRANCH Zero,je

op_F1:

; Illegal
op_F2:
; Illegal
op_F3:
; Illegal
op_F4:
	ret
	
op_F5:
op_F6:
op_F7:

; SED
op_F8:
	mov Decimal,1
	ADD_CYCLES 2
	inc _PC
	ret
	
op_F9:

; Illegal
op_FA:
; Illegal
op_FB:
; Illegal
op_FC:
	ret

op_FD:
op_FE:

; Illegal
op_FF:
	ret
	
;-------------------------------------------------------------------------------------------

	
opcode_table:
	dd OFFSET op_00
	dd OFFSET op_01
	dd OFFSET op_02
	dd OFFSET op_03
	dd OFFSET op_04
	dd OFFSET op_05
	dd OFFSET op_06
	dd OFFSET op_07
	dd OFFSET op_08
	dd OFFSET op_09
	dd OFFSET op_0A
	dd OFFSET op_0B
	dd OFFSET op_0C
	dd OFFSET op_0D
	dd OFFSET op_0E
	dd OFFSET op_0F

	dd OFFSET op_10
	dd OFFSET op_11
	dd OFFSET op_12
	dd OFFSET op_13
	dd OFFSET op_14
	dd OFFSET op_15
	dd OFFSET op_16
	dd OFFSET op_17
	dd OFFSET op_18
	dd OFFSET op_19
	dd OFFSET op_1A
	dd OFFSET op_1B
	dd OFFSET op_1C
	dd OFFSET op_1D
	dd OFFSET op_1E
	dd OFFSET op_1F

	dd OFFSET op_20
	dd OFFSET op_21
	dd OFFSET op_22
	dd OFFSET op_23
	dd OFFSET op_24
	dd OFFSET op_25
	dd OFFSET op_26
	dd OFFSET op_27
	dd OFFSET op_28
	dd OFFSET op_29
	dd OFFSET op_2A
	dd OFFSET op_2B
	dd OFFSET op_2C
	dd OFFSET op_2D
	dd OFFSET op_2E
	dd OFFSET op_2F

	dd OFFSET op_30
	dd OFFSET op_31
	dd OFFSET op_32
	dd OFFSET op_33
	dd OFFSET op_34
	dd OFFSET op_35
	dd OFFSET op_36
	dd OFFSET op_37
	dd OFFSET op_38
	dd OFFSET op_39
	dd OFFSET op_3A
	dd OFFSET op_3B
	dd OFFSET op_3C
	dd OFFSET op_3D
	dd OFFSET op_3E
	dd OFFSET op_3F

	dd OFFSET op_40
	dd OFFSET op_41
	dd OFFSET op_42
	dd OFFSET op_43
	dd OFFSET op_44
	dd OFFSET op_45
	dd OFFSET op_46
	dd OFFSET op_47
	dd OFFSET op_48
	dd OFFSET op_49
	dd OFFSET op_4A
	dd OFFSET op_4B
	dd OFFSET op_4C
	dd OFFSET op_4D
	dd OFFSET op_4E
	dd OFFSET op_4F

	dd OFFSET op_50
	dd OFFSET op_51
	dd OFFSET op_52
	dd OFFSET op_53
	dd OFFSET op_54
	dd OFFSET op_55
	dd OFFSET op_56
	dd OFFSET op_57
	dd OFFSET op_58
	dd OFFSET op_59
	dd OFFSET op_5A
	dd OFFSET op_5B
	dd OFFSET op_5C
	dd OFFSET op_5D
	dd OFFSET op_5E
	dd OFFSET op_5F

	dd OFFSET op_60
	dd OFFSET op_61
	dd OFFSET op_62
	dd OFFSET op_63
	dd OFFSET op_64
	dd OFFSET op_65
	dd OFFSET op_66
	dd OFFSET op_67
	dd OFFSET op_68
	dd OFFSET op_69
	dd OFFSET op_6A
	dd OFFSET op_6B
	dd OFFSET op_6C
	dd OFFSET op_6D
	dd OFFSET op_6E
	dd OFFSET op_6F

	dd OFFSET op_70
	dd OFFSET op_71
	dd OFFSET op_72
	dd OFFSET op_73
	dd OFFSET op_74
	dd OFFSET op_75
	dd OFFSET op_76
	dd OFFSET op_77
	dd OFFSET op_78
	dd OFFSET op_79
	dd OFFSET op_7A
	dd OFFSET op_7B
	dd OFFSET op_7C
	dd OFFSET op_7D
	dd OFFSET op_7E
	dd OFFSET op_7F

	dd OFFSET op_80
	dd OFFSET op_81
	dd OFFSET op_82
	dd OFFSET op_83
	dd OFFSET op_84
	dd OFFSET op_85
	dd OFFSET op_86
	dd OFFSET op_87
	dd OFFSET op_88
	dd OFFSET op_89
	dd OFFSET op_8A
	dd OFFSET op_8B
	dd OFFSET op_8C
	dd OFFSET op_8D
	dd OFFSET op_8E
	dd OFFSET op_8F

	dd OFFSET op_90
	dd OFFSET op_91
	dd OFFSET op_92
	dd OFFSET op_93
	dd OFFSET op_94
	dd OFFSET op_95
	dd OFFSET op_96
	dd OFFSET op_97
	dd OFFSET op_98
	dd OFFSET op_99
	dd OFFSET op_9A
	dd OFFSET op_9B
	dd OFFSET op_9C
	dd OFFSET op_9D
	dd OFFSET op_9E
	dd OFFSET op_9F

	dd OFFSET op_A0
	dd OFFSET op_A1
	dd OFFSET op_A2
	dd OFFSET op_A3
	dd OFFSET op_A4
	dd OFFSET op_A5
	dd OFFSET op_A6
	dd OFFSET op_A7
	dd OFFSET op_A8
	dd OFFSET op_A9
	dd OFFSET op_AA
	dd OFFSET op_AB
	dd OFFSET op_AC
	dd OFFSET op_AD
	dd OFFSET op_AE
	dd OFFSET op_AF

	dd OFFSET op_B0
	dd OFFSET op_B1
	dd OFFSET op_B2
	dd OFFSET op_B3
	dd OFFSET op_B4
	dd OFFSET op_B5
	dd OFFSET op_B6
	dd OFFSET op_B7
	dd OFFSET op_B8
	dd OFFSET op_B9
	dd OFFSET op_BA
	dd OFFSET op_BB
	dd OFFSET op_BC
	dd OFFSET op_BD
	dd OFFSET op_BE
	dd OFFSET op_BF

	dd OFFSET op_C0
	dd OFFSET op_C1
	dd OFFSET op_C2
	dd OFFSET op_C3
	dd OFFSET op_C4
	dd OFFSET op_C5
	dd OFFSET op_C6
	dd OFFSET op_C7
	dd OFFSET op_C8
	dd OFFSET op_C9
	dd OFFSET op_CA
	dd OFFSET op_CB
	dd OFFSET op_CC
	dd OFFSET op_CD
	dd OFFSET op_CE
	dd OFFSET op_CF

	dd OFFSET op_D0
	dd OFFSET op_D1
	dd OFFSET op_D2
	dd OFFSET op_D3
	dd OFFSET op_D4
	dd OFFSET op_D5
	dd OFFSET op_D6
	dd OFFSET op_D7
	dd OFFSET op_D8
	dd OFFSET op_D9
	dd OFFSET op_DA
	dd OFFSET op_DB
	dd OFFSET op_DC
	dd OFFSET op_DD
	dd OFFSET op_DE
	dd OFFSET op_DF

	dd OFFSET op_E0
	dd OFFSET op_E1
	dd OFFSET op_E2
	dd OFFSET op_E3
	dd OFFSET op_E4
	dd OFFSET op_E5
	dd OFFSET op_E6
	dd OFFSET op_E7
	dd OFFSET op_E8
	dd OFFSET op_E9
	dd OFFSET op_EA
	dd OFFSET op_EB
	dd OFFSET op_EC
	dd OFFSET op_ED
	dd OFFSET op_EE
	dd OFFSET op_EF

	dd OFFSET op_F0
	dd OFFSET op_F1
	dd OFFSET op_F2
	dd OFFSET op_F3
	dd OFFSET op_F4
	dd OFFSET op_F5
	dd OFFSET op_F6
	dd OFFSET op_F7
	dd OFFSET op_F8
	dd OFFSET op_F9
	dd OFFSET op_FA
	dd OFFSET op_FB
	dd OFFSET op_FC
	dd OFFSET op_FD
	dd OFFSET op_FE
	dd OFFSET op_FF	




cpu_init PROC commline:DWORD
	mov cycle,0

	mov eax,commline
	mov zombie,eax

	;invoke AllocConsole
	
	invoke GetStdHandle,STD_OUTPUT_HANDLE
	mov hout,eax
	
	ret
cpu_init ENDP


cpu_reset PROC
	pusha

	; Get MMU callbacks
	invoke zombie,ASK,MMU_READ_BYTE_CALLBACK,0,0
	mov mmu_read_byte,eax
	invoke zombie,ASK,MMU_READ_WORD_CALLBACK,0,0
	mov mmu_read_word,eax
	invoke zombie,ASK,MMU_WRITE_BYTE_CALLBACK,0,0
	mov mmu_write_byte,eax
	invoke zombie,ASK,MMU_WRITE_WORD_CALLBACK,0,0
	mov mmu_write_word,eax
	invoke zombie,ASK,MMU_TRANSLATE_ADDRESS_CALLBACK,0,0
	mov mmu_translate_address,eax

	; Reset registers
	mov edi,OFFSET start_regs
	mov ecx,OFFSET end_regs
	sub ecx,edi
	inc ecx
	mov eax,0
	rep stosb
	mov breakpoint,80000000h

	; Read reset vector
	mov ebx,0FFFCh
	invoke mmu_read_word,ebx
	mov _PC,ax
	
	popa
	ret
cpu_reset ENDP


cpu_close PROC
	;invoke FreeConsole
	ret
cpu_close ENDP


cpu_config PROC hWnd:DWORD
	ret
cpu_config ENDP


cpu_set_io_handler PROC mode:DWORD,port:DWORD,pfhandler:DWORD
	push ebx

	mov eax,port
	mov ebx,pfhandler
	
	.if mode==INPUT	
		mov [inpb_handlers + eax*4],ebx
	.else
		mov [outpb_handlers + eax*4],ebx
	.endif
	
	pop ebx
	
	ret
cpu_set_io_handler ENDP



cpu_get PROC what:DWORD
	.if what == CPU_CYCLE_COUNT
		mov eax,cycle
	.elseif what == CPU_SET_IO_CALLBACK
		mov eax,offset cpu_set_io_handler
	.elseif what == CPU_IRQ_CALLBACK
		mov eax,offset cpu_irq
	.elseif what == CPU_INPB_CALLBACK
		mov eax,offset cpu_inpb
	.elseif what == CPU_OUTPB_CALLBACK
		mov eax,offset cpu_outpb
	.elseif what == CPU_SPEED
		mov eax,speed
	.elseif what == CPU_CYCLE_SCALING
		mov eax,400h
	.elseif what == CPU_REG_PC
		movzx eax,_PC
	.elseif what == PLUGIN_NAME
		mov eax,offset cpuName
	.endif
	ret
cpu_get ENDP


cpu_set PROC what:DWORD,param1:DWORD,param2:DWORD
	.if what == CPU_REG_PC
		mov eax,param1
		mov _PC,ax
	.elseif what == CPU_REG_SP
		mov eax,param1
		mov _S,al
	.elseif what == CPU_CYCLE_COUNT
		mov eax,param1
		mov cycle,eax
	.elseif what == CPU_BREAKPOINT
		mov eax,param1
		mov breakpoint,eax
	.elseif what == CPU_SPEED
		mov eax,param1
		mov speed,eax
	.elseif what == CPU_IO_HANDLER
		mov eax,param1
		and param1,0FFh
		shr eax,8
		invoke cpu_set_io_handler,eax,param1,param2
	.endif
	ret
cpu_set ENDP



cpu_outpb PROC port:DWORD,dat:DWORD
	push ecx
	push dword ptr dat
	mov eax,port
	call dword ptr [outpb_handlers + eax*4]
	pop ecx
	ret
cpu_outpb ENDP



cpu_inpb PROC port:DWORD
	IFDEF _DEBUG_
		pusha
		invoke wsprintf, ADDR szBuffer,ADDR szFormat,port
		invoke MessageBox,NULL,ADDR szBuffer,ADDR szError,MB_ICONERROR
		popa
	ENDIF
	
	push ecx
	mov eax,port
	call dword ptr [inpb_handlers + eax*4]
	pop ecx
	ret
cpu_inpb ENDP

		
	
cpu_irq PROC address:DWORD
	.if Interrupt == 1
		pusha
		movzx eax,word ptr _PC
		PUSHWD eax
		call op_08	; PHP
		mov Interrupt,0
		mov eax,address
		mov _PC,ax
		add cycle,1C00h
		popa
	.endif
	ret
cpu_irq ENDP


cpu_execute PROC
	pusha
	
	mov ecx,cycle
	
	movzx eax,_PC
	push eax
	mov ebx,mmu_translate_address
	push eax
	call ebx
	mov esi,[eax+1]
	mov ebx,mmu_read_byte
	call ebx
	call dword ptr [opcode_table + eax*4]
	
	mov cycle,ecx
	popa
	mov eax,cycle
	ret
cpu_execute ENDP

	
cpu_execute_until PROC threshold:DWORD
	LOCAL lastCyc:DWORD
	
	pusha
	
	mov ecx,cycle
	
	@@exe_loop:
		cmp ecx,threshold
		jge @@exe_done
		;mov lastCyc,ecx

		movzx eax,_PC
		invoke mmu_translate_address,eax
		lea esi,[eax+1]
		
		movzx eax,_PC
		invoke mmu_read_byte,eax
		and eax,0FFh	
	
		call dword ptr [opcode_table + eax*4]
		mov ecx,cycle
	jmp @@exe_loop
	 
	@@exe_done:
	;mov cycle,ecx
	popa
	
	mov eax,cycle
	ret
cpu_execute_until ENDP


END LibMain
