# M*CORE fire effect for Zombie
# /Mic, 2005
#
# mcore-elf-as -cpu=210 -EB -o fire.o fire.s
# mcore-elf-ld -T lnkscript --oformat elf32-mcore-big fire.o
# mcore-elf-objcopy -O binary a.out fire.bin


# Address definitions for the fictive GPU
.EQU GPUREG0,	0xFF00
.EQU GPUREG1, 	0xFF01
.EQU VRAM, 	0x6000000
.EQU CRAM,	0x6012C00

# Misc. constants
.EQU READOFS, 	22718
.EQU WRITEOFS, 	22080
.EQU SCREENOFS,	75840


.TEXT
.ALIGN 2

.GLOBAL _start

_start:

	# Copy initialised data from ROM to RAM (132 bytes)
	lrw 	r1,rng_data_rom
	lrw 	r2,data_ram
	movi 	r3,33				# 33 = 132/4
	copy_data:
		ldw	r4,(r1)			# Load word from ROM
		stw	r4,(r2)			# Store word in RAM
		addi	r1,4			# Increase source pointer
		addi	r2,4			# Increase dest pointer
		decgt	r3			# Decrease R3 and set C if R3>0
		bt	copy_data		# Repeat until R3==0
		
	movi 	r1,0
	lrw 	r7,GPUREG0
	stb 	r1,(r7)				# Disable screen

	
	# Setup palette
	movi 	r8,0				# Loop counter
	movi 	r5,0x1F
	lrw 	r6,0xFFE0			# Color mask
	movi 	r1,64				# Used for address offsetting
	lrw 	r4,CRAM				# Color RAM address
	
	setup_palette:
		mov 	r2,r8
		lsli	r2,11
		mov	r10,r2			# Convert to little-endian
		zextb	r2			# ...	
		lsli	r2,8			# ...
		lsri	r10,8			# ...
		zextb	r10			# ...
		or	r2,r10			# ...
		sth 	r2,(r4)			# (r4) = r8 << 11
		
		addu 	r4,r1			# r4 += 64
		mov 	r3,r8
		mov 	r2,r5
		lsli	r3,6
		lsli	r2,11
		or 	r2,r3
		mov	r10,r2			# Convert to little-endian
		zextb	r2			# ...	
		lsli	r2,8			# ...
		lsri	r10,8			# ...
		zextb	r10			# ...
		or	r2,r10			# ...
		sth 	r2,(r4)			# (r4,64) = 0x1F<<11 | r8<<6
		
		mov 	r2,r6
		addu 	r4,r1			# r4 += 64
		or 	r2,r8
		mov	r10,r2			# Convert to little-endian
		zextb	r2			# ...	
		lsli	r2,8			# ...
		lsri	r10,8			# ...
		zextb	r10			# ...
		or	r2,r10			# ...
		sth 	r2,(r4)			# (r4,128) = 0xFFE0 | r8
		subu 	r4,r1
		subu 	r4,r1
		
		addi 	r4,2
		addi 	r8,1
		movi 	r9,0x20
		cmpne 	r8,r9
		bt 	setup_palette
		
		
	# Set the remaining colors to white
	lrw	r4,CRAM
	lrw	r5,192

	addu	r4,r5
	lrw	r8,0xFFFF			# 0xFFFF == full white
	lrw	r1,160

	set_colors_to_white:
		sth	r8,(r4)
		addi 	r4,2
		decgt	r1
		bt set_colors_to_white

	lrw	r8,0xB4
	stb	r8,(r7)				# Enable display, vblank interrupts, bitmapped mode


	forever:
		# Wait for vertical blank
		lrw	r5,GPUREG1
		lrw	r6,0x80
		movi	r9,0x40
		lslc	r9
		wait_vb:
			ldb	r1,(r5) 
			and	r1,r9		# Check vblank bit
			cmpne 	r1,r6
			bt 	wait_vb
			
	
		movi r8,0			# X
		x_loop:
			lrw	r4,VRAM
			addu	r4,r8
			lrw	r5,READOFS
			lrw	r6,WRITEOFS
			addu	r5,r4
			addu	r6,r4
			movi	r9,71		# Start at line 71
			y_loop:
				lrw 	r7,640	# Size of two scanlines, in bytes
				addi	r7,4
				addu	r7,r5
				
				# Read 4 pixels
				ldb	r10,(r5)
				ldb	r1,(r5,2)
				mov	r11,r1
				ldb	r1,(r5,4)
				mov	r12,r1
				ldb	r13,(r7)

				# Clear upper 24 bits of r10..r13
				zextb	r10
				zextb	r11
				zextb	r12
				zextb	r13
				
				# Calculate average
				addu	r10,r11
				addu	r10,r12
				addu	r10,r13
				mov 	r1,r10
				lsri	r1,2
				
				# Decrement color value if >0
				cmpnei 	r1,0
				dect 	r1

				# Duplicate bits 0..7 into bits 8..15
				mov	r10,r1
				lsli	r10,8
				or	r1,r10
				
				lrw	r4,320
				add	r4,r6
				sth	r1,(r6)	# Draw 2 pixels
				sth	r1,(r4)	# Draw 2 more pixels

				lrw	r10,640
				addu	r5,r10	# Move 2 rows down
				addu	r6,r10	# Dito
				
				lrw	r11,239
				addi	r9,2
				cmpne	r9,r11
				bt	y_loop

			bsr	Random

			addi	r2,28
			lrw 	r10,191
			and	r2,r10
			
			lrw	r4,VRAM
			lrw	r11,SCREENOFS
			addu	r4,r8		# Add X
			addu	r4,r11
			stb	r2,(r4)		# Store random pixel

			addi	r8,2
			lrw	r10,320
			cmpne	r8,r10
			bt	x_loop
	
		br forever			


	# Pseudo-random number generator
	Random:
		lrw	r10,data_ram		# Point to seed1 in RAM
		mov	r3,r10
		addi	r3,2			# Point to the random number table
		ldb	r11,(r10)		# r11 = seed1
		ldb	r1,(r10,1)		# r1  = seed2
		mov	r12,r11
		mov	r13,r1
		lslc	r12
		lslc	r13
		addu	r12,r3
		addu	r13,r3
		ldh	r12,(r12)		# Load random number
		ldh	r14,(r13)		# ...another one
		addu	r12,r14			# Add the two numbers
		sth	r12,(r13)		# Store it
		addi	r11,1			# Increase seed1
		addi	r1,1			# Increase seed2
		movi	r14,0x3F		# The table contains 64 entries
		and	r11,r14
		and	r1,r14
		mov	r2,r12			# Put the random number in R2
		stb	r11,(r10)
		stb	r1,(r10,1)
		jmp	r15			# Return		

.align 4


rng_data_rom:	
	# Seeds
	.byte 29
	.byte 63

	# 64 random 16-bit numbers for use with the random-number generator 
	.short 25662,56977,36314,46707
	.short 56747,47936,28467,50155
	.short 54470,9126,19191,26274
	.short 27205,13469,6769,48786
	.short 43195,17969,58549,13097
	.short 9962,26613,60724,25428
	.short 44416,43884,31777,40699
	.short 4280,32767,31882,41912
	.short 19050,42386,38267,39411
	.short 11739,36780,10921,43911
	.short 7473,38901,28614,62047
	.short 25094,22064,16149,14666
	.short 46209,34643,30017,25842
	.short 64242,3527,34102,35358
	.short 2277,3678,16479,3431
	.short 36135,63741,11311,0
          

.DATA
data_ram:
	seed1: .byte
	seed2: .byte
	random_tbl:
	