// Customizable Z80 MMU emulator
// Mic, 2004

#include <windows.h>
#include <commctrl.h>
#include <string.h>
#include "resrc1.h"
#include "zombie.h"


#define JUMP 0

HINSTANCE hInst;
HANDLE hout;
unsigned char **pageTable,*RAM,*cart;
unsigned char *iomap;
int numPages=2,pageSize=0x4000,pageRegBase;
int ramOffset,ramSize=0x8000;
char mmuName[] = "Z80 MMU";
int dummy;
cpu_inpb_callback cpu_inpb;
cpu_outpb_callback cpu_outpb;
zombie_callback zombie;


BOOL CALLBACK DlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam);



bool __stdcall DllMain(HINSTANCE hDLL,DWORD reason,LPVOID unused)
{
	if (reason==DLL_PROCESS_ATTACH)
		hInst = hDLL;

	return true;
}


/*char* __stdcall mmu_get_name()
{
	return mmuName;
}*/

bool __stdcall mmu_init(zombie_callback commlink)
{
	zombie = commlink;

	ramOffset = pageSize*numPages;
	ramSize = 0x10000-ramOffset;
	pageRegBase = (0x10000-numPages)-ramOffset;

	RAM = new unsigned char[ramSize];
	pageTable = new unsigned char*[numPages];
	iomap = new unsigned char[0x10000];

	for (int i=0; i<0x10000; i++)
		iomap[i] = 0xFF;

	hout = GetStdHandle(STD_OUTPUT_HANDLE);
	return true;
}


bool __stdcall mmu_reset()
{
	cpu_inpb = (cpu_inpb_callback)zombie(ASK,CPU_INPB_CALLBACK,0,0);
	cpu_outpb = (cpu_outpb_callback)zombie(ASK,CPU_OUTPB_CALLBACK,0,0);

	cart = (unsigned char*)zombie(ASK,ZOMBIE_ROM_BASE,0,0);

	zombie(TELL,GPU_VBLANK_ACTION,(void*)JUMP,(void*)0x38);
	zombie(TELL,GPU_HBLANK_ACTION,(void*)JUMP,(void*)0x38);

	memset(RAM,0,ramSize);
	for (int i=0; i<numPages; i++)
		pageTable[i] = cart+i*pageSize;

	return true;
}


void __stdcall mmu_close()
{
	delete [] pageTable;
	delete [] RAM;
}



int __stdcall mmu_config(HWND hwnd)
{
	return DialogBoxParam(hInst,
            MAKEINTRESOURCE(IDD_MMU_CONFIG), hwnd,
            DlgProc, NULL);
}




unsigned char* __stdcall mmu_translate_address(int address)
{
	if (address<ramOffset)
			//return pageTable[address/pageSize]+(address&(pageSize-1));
			return pageTable[address>>14]+(address&(pageSize-1));

	return &RAM[address&(ramSize-1)];
}



unsigned char __stdcall mmu_read_byte(int address)
{
	static char buffer[64];

	if (iomap[address] != 0xFF)
		return cpu_inpb(iomap[address]);

	if (address<ramOffset)
		//return *(pageTable[address/pageSize]+(address&(pageSize-1)));
		return *(pageTable[address>>14]+(address&(pageSize-1)));

	return RAM[address&(ramSize-1)];
}


short int __stdcall mmu_read_word(int address)
{
	return mmu_read_byte(address)+(mmu_read_byte(address+1)<<8);
}


void __stdcall mmu_write_byte(int address,int byte)
{
	static char buffer[64];

	if (iomap[address] != 0xFF)
		cpu_outpb(iomap[address],byte);

	if (address>=ramOffset)
	{
		address &= ramSize-1;
		RAM[address] = (unsigned char)byte;

		if (address>=pageRegBase)
			pageTable[address-pageRegBase] = cart+(byte&numPages-1)*pageSize;
	}
}


void __stdcall mmu_write_word(int address,int word)
{
	mmu_write_byte(address,word&0xFF);
	mmu_write_byte(address+1,word>>8);
}


int __stdcall mmu_get(int what)
{
	switch (what)
	{
		case MMU_READ_BYTE_CALLBACK:
			return (int)mmu_read_byte;
		case MMU_READ_WORD_CALLBACK:
			return (int)mmu_read_word;
		case MMU_WRITE_BYTE_CALLBACK:
			return (int)mmu_write_byte;
		case MMU_WRITE_WORD_CALLBACK:
			return (int)mmu_write_word;
		case MMU_TRANSLATE_ADDRESS_CALLBACK:
			return (int)mmu_translate_address;
		case PLUGIN_NAME:
			return (int)&mmuName;
		default:
			break;
	}
	return 0;
}


void __stdcall mmu_set(int what,void* param1,void* param2)
{
	static char buffer[64];

	switch (what)
	{
		case MMU_MAP_PORT:
			iomap[(int)param2] = (unsigned char)param1;
			//wsprintf(buffer,"Mapping port %02x to address %04x\n",iomap[(int)param2],param2);
			//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long *)&dummy,NULL);
			//MessageBox(NULL,buffer,"mmu_set",MB_OK);
			break;
		case 999:
			cpu_inpb(0);
			//inpb_00();

		default:
			break;
	}
}


//////////////////////////////////// Configuration //////////////////////////////////////////////

void tbm_setrange(HWND hdlg,int id,int low,int hi)
{
	SendMessage(GetDlgItem(hdlg,id),TBM_SETRANGE,(WPARAM)true,(LPARAM)MAKELONG(low,hi));
}

void tbm_setpos(HWND hdlg,int id,int pos)
{
	SendMessage(GetDlgItem(hdlg,id),TBM_SETPOS,(WPARAM)true,(LPARAM)pos);
}

int tbm_getpos(HWND hdlg,int id)
{
	return (int)SendMessage(GetDlgItem(hdlg,id),TBM_GETPOS,(WPARAM)0,(LPARAM)0);
}



BOOL CALLBACK DlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam) {
	static char stemp[64];

    switch(msg) {
		case WM_INITDIALOG:
			SetWindowLong(hdlg, DWL_USER, lParam);

			tbm_setrange(hdlg,IDC_RAMSLIDER,1,32);
			tbm_setpos(hdlg,IDC_RAMSLIDER,ramSize>>10);
			wsprintf(stemp,"%d kB ",ramSize>>10);
			SetWindowText(GetDlgItem(hdlg,IDC_RAMTEXT),stemp);

			tbm_setrange(hdlg,IDC_NPAGESLIDER,1,8);
			tbm_setpos(hdlg,IDC_NPAGESLIDER,numPages);
			wsprintf(stemp,"%d ",numPages);
			SetWindowText(GetDlgItem(hdlg,IDC_NPAGETEXT),stemp);

			tbm_setrange(hdlg,IDC_PAGESIZESLIDER,1,48);
			tbm_setpos(hdlg,IDC_PAGESIZESLIDER,pageSize>>10);
			wsprintf(stemp,"%d kB ",pageSize>>10);
			SetWindowText(GetDlgItem(hdlg,IDC_PAGESIZETEXT),stemp);

			wsprintf(stemp,"%04x",pageRegBase+ramOffset);
			SetWindowText(GetDlgItem(hdlg,IDC_PAGEEDIT),stemp);

			return TRUE;

		case WM_HSCROLL:
			if ((HWND)lParam==GetDlgItem(hdlg,IDC_RAMSLIDER))
			{
				ramSize = tbm_getpos(hdlg,IDC_RAMSLIDER)<<10;
				wsprintf(stemp,"%d kB ",ramSize>>10);
				SetWindowText(GetDlgItem(hdlg,IDC_RAMTEXT),stemp);

				if (ramSize+(numPages*pageSize)>0x10000)
				{
					pageSize = ((0x10000-ramSize)/numPages)>>10;
					tbm_setpos(hdlg,IDC_PAGESIZESLIDER,pageSize>>10);
					wsprintf(stemp,"%d kB ",pageSize);
					SetWindowText(GetDlgItem(hdlg,IDC_PAGESIZETEXT),stemp);
					pageSize <<= 10;
				}
			} else
			if ((HWND)lParam==GetDlgItem(hdlg,IDC_NPAGESLIDER))
			{
				numPages = tbm_getpos(hdlg,IDC_NPAGESLIDER);
				wsprintf(stemp,"%d ",numPages);
				SetWindowText(GetDlgItem(hdlg,IDC_NPAGETEXT),stemp);

				if (ramSize+(numPages*pageSize)>0x10000)
				{
					pageSize = ((0x10000-ramSize)/numPages)>>10;
					tbm_setpos(hdlg,IDC_PAGESIZESLIDER,pageSize>>10);
					wsprintf(stemp,"%d kB ",pageSize);
					SetWindowText(GetDlgItem(hdlg,IDC_PAGESIZETEXT),stemp);
					pageSize <<= 10;
				}
			} else
			if ((HWND)lParam==GetDlgItem(hdlg,IDC_PAGESIZESLIDER))
			{
				pageSize = tbm_getpos(hdlg,IDC_PAGESIZESLIDER)<<10;
				wsprintf(stemp,"%d kB ",numPages);
				SetWindowText(GetDlgItem(hdlg,IDC_PAGESIZETEXT),stemp);

				if (ramSize+(numPages*pageSize)>0x10000)
				{
					pageSize = ((0x10000-ramSize)/numPages)>>10;
					tbm_setpos(hdlg,IDC_PAGESIZESLIDER,pageSize>>10);
					wsprintf(stemp,"%d kB ",pageSize);
					SetWindowText(GetDlgItem(hdlg,IDC_PAGESIZETEXT),stemp);
					pageSize <<= 10;
				}
			}
			break;

		case WM_COMMAND:
			switch(LOWORD(wParam)) {
				case IDOK:
					delete [] RAM;
					RAM = new unsigned char[ramSize];

					EndDialog(hdlg, 0);
					return TRUE;

				case IDCANCEL:
					EndDialog(hdlg, 1);
					return FALSE;

				default:
					break;
			}
		break;
	}
    return FALSE;
}

