; Zombie test
; Uses tiled mode (320x240x8bpp) and sprites
; /Mic 2004
;
; Assemble with WLA-DX


.define PAGE_SIZE $4000

.memorymap
	defaultslot 0

	slotsize PAGE_SIZE
	slot 0 0
	slot 1 PAGE_SIZE
	
	slotsize $8000
	slot 2 $8000
.endme


.rombanksize $4000
.rombanks 2


.bank 0
.org $0000


di
im 1
ld sp,$dff0				; Set stack pointer
jp main


; IRQ handler
.org $0038
	di
	jp vblank_routine


; NMI handler	
.org $0066
	retn


.org $0100

main:
	ld a,$04			; Tiled mode, 8bpp
	ld ($FF00),a

	; Set VRAM pointer to $12DE0 (palette[241])
	ld a,$E2
	ld ($FF10),a			; $FF10 = GPU address register
	ld a,$2D
	ld ($FF10),a
	ld a,$01
	ld ($FF10),a
	ld a,$00
	ld ($FF10),a
	
	ld de,$FF11			; $FF11 = GPU data register
	ld hl,palette+PAGE_SIZE
	ld b,6				; 3 colors = 6 bytes
	set_palette:
		ld a,(hl)
		ld (de),a
		inc hl
		dec b
		jr nz,set_palette

	; Point to start of pattern table
	ld a,$00
	ld ($FF10),a
	ld ($FF10),a
	ld ($FF10),a
	ld ($FF10),a
	
	ld hl,tiles+PAGE_SIZE
	ld b,0
	ld c,32				; Load 32 tiles
	load_tiles:
		ld a,(hl)
		ld ($FF11),a
		inc hl
		dec b
		jr nz,load_tiles
		dec c
		jr nz,load_tiles


	; Set VRAM pointer to $12E00 (sprite memory)
	ld a,$00
	ld ($FF10),a
	ld a,$2E
	ld ($FF10),a
	ld a,$01
	ld ($FF10),a
	ld a,$00
	ld ($FF10),a
	ld a,$FF			; The idea here is to set the y-coordinate of
	ld b,0				; every sprite to $FF (ie. put it off screen)
	clear_sprites:
		ld ($FF11),a
		dec b
		jr nz,clear_sprites


	; Print some text on the background		
	ld a,$00
	ld ($FF10),a
	ld a,$01
	ld ($FF10),a
	ld ($FF10),a
	ld a,$00
	ld ($FF10),a
	
	ld hl,string1+PAGE_SIZE
	print_string1:
		ld a,(hl)
		cp $00
		jr z,print_string1_done
		sub 64			; Subtract 64 = '@'
		ld ($FF11),a		; Store the tile number
		ld a,$00
		ld ($FF11),a		; Set the high byte of the name table entry
		inc hl			; Next character
		dec b
		jp print_string1
	print_string1_done:
	
	ld a,$A4
	ld ($FF00),a			; Enable vblank, enable display, use tiled mode, 8bpp
	
	ei				; Enable interrupts

	forever:
		jp forever		; Sit and wait for interrupts





vblank_routine:
	ld a,($FF01)			; Read GPU register 1, vblank flag is cleared
	and $80
	jr z,hblank

	; Point to sprite memory
	ld a,$00
	ld ($FF10),a
	ld a,$2E
	ld ($FF10),a
	ld a,$01
	ld ($FF10),a
	ld a,$00
	ld ($FF10),a

	ld hl,string2+PAGE_SIZE
	ld de,sine+PAGE_SIZE
	ld a,(angle)
	add a,e
	ld e,a
	ld c,80
	print_string2:
		ld a,(hl)		; Load character
		cp $00
		jr z,print_string2_done
		sub 64
		ld b,a
		
		ld a,(de)		; Load y-coordinate from sine table
		add a,$80		; Offset by 128
		ld ($FF11),a
		
		ld a,c
		add a,$10
		ld ($FF11),a		; Set x-coordinate
		ld c,a
		
		sll b			; Multiply tile number by 2
		ld a,b
		ld ($FF11),a		; Set tile number
		
		ld a,$00
		ld ($FF11),a
		
		inc hl			; Next character
		inc e			; Increase sine table pointer
		dec b
		jp print_string2
	print_string2_done:
	
	ld a,(angle)
	inc a
	ld (angle),a
	
	hblank:
	ei
	reti
	
	
.bank 1
.org $0000

sine:
	.incbin "sine.bin"
tiles:
	.incbin "serif8.bin"
palette:
	.dw $083E,$7BDF,$CEFF
string1:
	.db "@@@@@BACKGROUND"
	.db 0
string2:
	.db "SPRITES"
	.db 0


.enum $8000

angle db

.ende



	