# MIPS R2000 rotozoomer for Zombie
# /Mic, 2004
#
# Assemble with ee-as (http://www.lukasz.dk/ps2toolchain.html)


.TEXT
.ALIGN

# Constants
.EQU GPU_REG0,0xFF00
	.EQU VBLANK_ENABLE,0x80
	.EQU HBLANK_ENABLE,0x40

	.EQU DISPLAY_ENABLE,0x20
	.EQU BITMAPPED_MODE,0x10
	.EQU COLOR_8BIT,0x04
	
.EQU GPU_REG1,0xFF01
	.EQU VBLANK_BIT,0x80

.EQU EXRAM,0x2000000				# External RAM (I'm using an ARM-style MMU for this)
.EQU SCREEN,0x6000000				# Start of VRAM
.EQU PALETTE,0x6012C00				# Start of palette RAM
.EQU SPRITE_TABLE,0x6012E00			# Start of sprite RAM


.GLOBAL _start
.ENT _start

_start:

	ori $23,$0,1				# Used as a loop decrement

	# Disable display		
	ori $8,$0,0
	ori $7,$0,GPU_REG0
	sb $8,($7)
	

	# Unpack texture into RAM
	la $4,suupaman_rle
	li $5,EXRAM
	li $6,0x10000
	jal Unpack


	# Load palette
	and $4,$4,$0
	and $5,$5,$0
	la $4,suupaman_pal
	li $5,PALETTE
	ori $8,$0,0x80				# 0x80*4 = 512 byte
	load_palette:
		lw $10,($4)
		sw $10,($5)
		addiu $4,$4,4
		addiu $5,$5,4
		subu $8,$8,$23
		bne $8,$0,load_palette


	# Move all sprites off-screen
	and $5,$5,$0
	li $5,SPRITE_TABLE
	ori $8,$0,0x80				# 128 sprites
	ori $9,$0,0xFF				
	set_sprites:
		sw $9,($5)
		addiu $5,$5,4
		subu $8,$8,$23
		bne $8,$0,set_sprites
		

	# Copy initialised variables from ROM to RAM
	and $4,$4,$0
	and $5,$5,$0
	la $4,init_data_start
	la $5,rotX
	ori $8,$0,6
	copy_data:
		lw $10,($4)
		sw $10,($5)
		addiu $4,$4,4
		addiu $5,$5,4
		subu $8,$8,$23
		bne $8,$0,copy_data
	
	
	# Enable display, set 320x240 8bpp bitmapped mode		
	ori $8,$0,VBLANK_ENABLE | DISPLAY_ENABLE | BITMAPPED_MODE | COLOR_8BIT
	ori $7,$0,GPU_REG0
	sb $8,($7)

	

	la $16,sine
	la $17,rotX

	forever:
		# Wait for vertical blank
		ori $5,$0,GPU_REG1
		ori $9,$0,VBLANK_BIT
		wait_vb:
			lbu $8,($5)
			and $8,$8,$9
			bne $8,$9,wait_vb
		
		lw $4,($17)			# rotX
		lw $5,4($17)			# rotY
		lw $6,20($17)			# distance
		lw $7,8($17)			# rot2
		jal Draw
		
		lw $8,8($17)
		addiu $8,$8,2
		andi $8,$8,0xFF
		sw $8,8($17)			# rot2 = (rot2 + speed) & 0xFF
		
		lw $8,4($17)
		addiu $8,$8,128
		sw $8,4($17)			# rotY += 128
		
		lw $8,20($17)
		lw $9,16($17)
		addu $8,$8,$9
		andi $8,$8,0xFFFF		# distance = (distance + speedZ) & 0xFFFF
		
		ori $2,$0,1400
		subu $3,$8,$2
		bgtz $3,turn			# Branch if distance>2400
		ori $2,$0,100
		subu $3,$2,$8
		bgtz $3,turn			# Branch if distance<3
		j no_turn
		turn:
			xori $9,$9,0xFFFF	# Invert speedZ
		no_turn:
		sw $8,20($17)
		sw $9,16($17)
		
		j forever



	# $4 : x
	# $5 : y
	# $6 : scale
	# $7 : rot
	Draw:
		addiu $8,$7,64
		andi $8,$8,0xFF
		sll $7,$7,2			# Shift left by 2 for word addressing
		sll $8,$8,2
		addu $18,$16,$7			# sin
		addu $19,$16,$8			# cos
		
		lw $10,($18)			# $10 = sin(rot)
		lw $11,($19)			# $11 = cos(rot)
		multu $10,$6
		mflo $12
		multu $11,$6
		mflo $13
		sra $12,$12,8			# ddy = (sin(rot)*scale)>>8
		sra $13,$13,8			# ddx = (cos(rot)*scale)>>8)

		or $14,$13,$0			# d2y = ddx
		subu $15,$0,$12			# d2x = -ddy
		
		ori $6,$0,160
		ori $7,$0,120

		multu $13,$6			# ddx*160
		mflo $10
		multu $15,$7			# d2x*120
		mflo $11
		subu $4,$4,$10
		subu $4,$4,$11			# x - ddx*160 - d2x*120

		multu $12,$6			# ddy*160
		mflo $10
		multu $14,$7			# d2y*120
		mflo $11
		subu $5,$5,$10
		subu $5,$5,$11			# y - ddy*160 - d2y*120

		or $10,$4,$0
		or $11,$5,$0


	# $10 : i
	# $11 : j
	# $12 : ddy
	# $13 : ddx
	# $14 : d2y
	# $15 : d2x
	Zoom:
		#ori $10,$0,0
		#ori $11,$0,0
		
		ori $4,$0,0
		ori $5,$0,0
		li $4,0x2000000			# Source
		li $5,SCREEN			# Dest

		ori $24,$0,240			# Y counter
		y_loop:
			or $2,$10,$0		# i
			or $3,$11,$0		# j
			ori $25,$0,320		# X counter
			x_loop:
				addu $2,$2,$13
				addu $3,$3,$12
				srl $8,$2,8
				andi $9,$3,0xFF00
				andi $8,$8,0xFF
				or $8,$8,$9
				addu $7,$4,$8
				
				lb $6,($7)	# Load texel
				sb $6,($5)	# Store
				addiu $5,$5,1	# Increment dest address
				
				subu $25,$25,$23
				bne $25,$0,x_loop
			addu $10,$10,$15
			addu $11,$11,$14
			subu $24,$24,$23
			bne $24,$0,y_loop
		jr $31
			



	# $4 : src pointer
	# $5 : dest pointer
	# $6 : Length of dest stream
	.EQU TRIGGER,0xFE
	Unpack:
		ori $24,$0,TRIGGER
		ori $8,$0,0
		ori $9,$0,0
		ori $10,$0,0
		ori $11,$0,1
		and $2,$2,$0			# Bytes unpacked
		unpack_loop:
			lbu $8,($4)
			bne $8,$24,single_byte
				ori $9,$0,0
				lbu $9,1($4)
				bne $9,$0,unpack_run
					addiu $4,$4,2
					sb $24,($5)
					addiu $5,$5,1
					addiu $2,$2,1
					j unpack_next
				unpack_run:
					addiu $9,$9,1
					lbu $10,2($4)
					addu $2,$2,$9
					unpack_run_inner:
						sb $10,($5)
						addiu $5,$5,1
						subu $9,$9,$11
						bne $9,$0,unpack_run_inner
					addiu $4,$4,3
					j unpack_next
				single_byte:
					sb $8,($5)
					addiu $4,$4,1
					addiu $5,$5,1
					addiu $2,$2,1
			unpack_next:
				subu $12,$6,$2
				blez $12,unpack_done
				j unpack_loop
		
			unpack_done:
				jr $31
		
init_data_start:
.WORD 32768
.WORD 0
.WORD 1
.WORD 2
.WORD 65522
.WORD 1000


sine:	# for (i=0; i<256; i++) (int)sin(i*PI/128)*256;
.WORD 0x00000000,0x00000006,0x0000000C,0x00000012,0x00000019,0x0000001F,0x00000025,0x0000002B
.WORD 0x00000031,0x00000038,0x0000003E,0x00000044,0x0000004A,0x00000050,0x00000056,0x0000005C
.WORD 0x00000061,0x00000067,0x0000006D,0x00000073,0x00000078,0x0000007E,0x00000083,0x00000088
.WORD 0x0000008E,0x00000093,0x00000098,0x0000009D,0x000000A2,0x000000A7,0x000000AB,0x000000B0
.WORD 0x000000B5,0x000000B9,0x000000BD,0x000000C1,0x000000C5,0x000000C9,0x000000CD,0x000000D1
.WORD 0x000000D4,0x000000D8,0x000000DB,0x000000DE,0x000000E1,0x000000E4,0x000000E7,0x000000EA
.WORD 0x000000EC,0x000000EE,0x000000F1,0x000000F3,0x000000F4,0x000000F6,0x000000F8,0x000000F9
.WORD 0x000000FB,0x000000FC,0x000000FD,0x000000FE,0x000000FE,0x000000FF,0x000000FF,0x000000FF
.WORD 0x00000100,0x000000FF,0x000000FF,0x000000FF,0x000000FE,0x000000FE,0x000000FD,0x000000FC
.WORD 0x000000FB,0x000000F9,0x000000F8,0x000000F6,0x000000F4,0x000000F3,0x000000F1,0x000000EE
.WORD 0x000000EC,0x000000EA,0x000000E7,0x000000E4,0x000000E1,0x000000DE,0x000000DB,0x000000D8
.WORD 0x000000D4,0x000000D1,0x000000CD,0x000000C9,0x000000C5,0x000000C1,0x000000BD,0x000000B9
.WORD 0x000000B5,0x000000B0,0x000000AB,0x000000A7,0x000000A2,0x0000009D,0x00000098,0x00000093
.WORD 0x0000008E,0x00000088,0x00000083,0x0000007E,0x00000078,0x00000073,0x0000006D,0x00000067
.WORD 0x00000061,0x0000005C,0x00000056,0x00000050,0x0000004A,0x00000044,0x0000003E,0x00000038
.WORD 0x00000031,0x0000002B,0x00000025,0x0000001F,0x00000019,0x00000012,0x0000000C,0x00000006
.WORD 0x00000000,0xFFFFFFF9,0xFFFFFFF3,0xFFFFFFED,0xFFFFFFE6,0xFFFFFFE0,0xFFFFFFDA,0xFFFFFFD4
.WORD 0xFFFFFFCE,0xFFFFFFC7,0xFFFFFFC1,0xFFFFFFBB,0xFFFFFFB5,0xFFFFFFAF,0xFFFFFFA9,0xFFFFFFA3
.WORD 0xFFFFFF9E,0xFFFFFF98,0xFFFFFF92,0xFFFFFF8C,0xFFFFFF87,0xFFFFFF81,0xFFFFFF7C,0xFFFFFF77
.WORD 0xFFFFFF71,0xFFFFFF6C,0xFFFFFF67,0xFFFFFF62,0xFFFFFF5D,0xFFFFFF58,0xFFFFFF54,0xFFFFFF4F
.WORD 0xFFFFFF4A,0xFFFFFF46,0xFFFFFF42,0xFFFFFF3E,0xFFFFFF3A,0xFFFFFF36,0xFFFFFF32,0xFFFFFF2E
.WORD 0xFFFFFF2B,0xFFFFFF27,0xFFFFFF24,0xFFFFFF21,0xFFFFFF1E,0xFFFFFF1B,0xFFFFFF18,0xFFFFFF15
.WORD 0xFFFFFF13,0xFFFFFF11,0xFFFFFF0E,0xFFFFFF0C,0xFFFFFF0B,0xFFFFFF09,0xFFFFFF07,0xFFFFFF06
.WORD 0xFFFFFF04,0xFFFFFF03,0xFFFFFF02,0xFFFFFF01,0xFFFFFF01,0xFFFFFF00,0xFFFFFF00,0xFFFFFF00
.WORD 0xFFFFFF00,0xFFFFFF00,0xFFFFFF00,0xFFFFFF00,0xFFFFFF01,0xFFFFFF01,0xFFFFFF02,0xFFFFFF03
.WORD 0xFFFFFF04,0xFFFFFF06,0xFFFFFF07,0xFFFFFF09,0xFFFFFF0B,0xFFFFFF0C,0xFFFFFF0E,0xFFFFFF11
.WORD 0xFFFFFF13,0xFFFFFF15,0xFFFFFF18,0xFFFFFF1B,0xFFFFFF1E,0xFFFFFF21,0xFFFFFF24,0xFFFFFF27
.WORD 0xFFFFFF2B,0xFFFFFF2E,0xFFFFFF32,0xFFFFFF36,0xFFFFFF3A,0xFFFFFF3E,0xFFFFFF42,0xFFFFFF46
.WORD 0xFFFFFF4A,0xFFFFFF4F,0xFFFFFF54,0xFFFFFF58,0xFFFFFF5D,0xFFFFFF62,0xFFFFFF67,0xFFFFFF6C
.WORD 0xFFFFFF71,0xFFFFFF77,0xFFFFFF7C,0xFFFFFF81,0xFFFFFF87,0xFFFFFF8C,0xFFFFFF92,0xFFFFFF98
.WORD 0xFFFFFF9E,0xFFFFFFA3,0xFFFFFFA9,0xFFFFFFAF,0xFFFFFFB5,0xFFFFFFBB,0xFFFFFFC1,0xFFFFFFC7
.WORD 0xFFFFFFCE,0xFFFFFFD4,0xFFFFFFDA,0xFFFFFFE0,0xFFFFFFE6,0xFFFFFFED,0xFFFFFFF3,0xFFFFFFF9


.END _start


.DATA
rotX: .WORD
rotY: .WORD
rot2: .WORD
speed: .WORD
speedZ: .WORD
distance: .WORD
